@@ -1,5 +1,5 @@
 /* hdparm.c - Command line interface to get/set hard disk parameters */
-/*          - by Mark S. Lord (C) 1994-2000 -- freely distributable */
+/*          - by Mark S. Lord (C) 1994-2001 -- freely distributable */
 
 #include <unistd.h>
 #include <string.h>
@@ -16,10 +16,10 @@
 #include <sys/times.h>
 #include <sys/types.h>
 #include <linux/hdreg.h>
-#include <linux/fs.h>
 #include <linux/major.h>
+#include <sys/mount.h>
 
-#define VERSION "v3.9"
+#define VERSION "v4.1"
 
 #undef	 DO_FLUSHCACHE		/* under construction: force cache flush on -W0 */
 
@@ -73,13 +73,14 @@ static unsigned long set_standby  = 0, get_standby  = 0, standby  = 0;
 static unsigned long set_xfermode = 0, get_xfermode = 0, xfermode = 0;
 static unsigned long set_lookahead= 0, get_lookahead= 0, lookahead= 0;
 static unsigned long set_prefetch = 0, get_prefetch = 0, prefetch = 0;
-static unsigned long set_defects  = 0, defects = 0;
+static unsigned long set_defects  = 0, get_defects  = 0, defects  = 0;
 static unsigned long set_wcache   = 0, get_wcache   = 0, wcache   = 0;
 static unsigned long set_doorlock = 0, get_doorlock = 0, doorlock = 0;
 static unsigned long set_seagate  = 0, get_seagate  = 0;
 static unsigned long set_standbynow = 0, get_standbynow = 0;
 static unsigned long set_sleepnow   = 0, get_sleepnow   = 0;
 static unsigned long get_powermode  = 0;
+static unsigned long set_apmmode = 0, get_apmmode= 0, apmmode = 0;
 #endif
 #ifdef CDROM_SELECT_SPEED
 static unsigned long set_cdromspeed = 0, cdromspeed = 0;
@@ -98,13 +99,63 @@ static int	hwif_ctrl = 0;
 static int	hwif_irq = 0;
 #endif
 
+#ifdef HDIO_DRIVE_RESET
+static int	perform_reset = 0;
+#endif /* HDIO_DRIVE_RESET */
+#ifdef HDIO_TRISTATE_HWIF
+static int	perform_tristate = 0,	tristate = 0;
+#endif /* HDIO_TRISTATE_HWIF */
+
 char *cfg_str[] =
-{	"", " HardSect", " SoftSect", " NotMFM", " HdSw>15uSec", " SpinMotCtl",
-	" Fixed", " Removeable", " DTR<=5Mbs", " DTR>5Mbs", " DTR>10Mbs",
-	" RotSpdTol>.5%", " dStbOff", " TrkOff", " FmtGapReq", " nonMagnetic"
+{	"",             " HardSect",   " SoftSect",  " NotMFM",
+	" HdSw>15uSec", " SpinMotCtl", " Fixed",     " Removeable",
+	" DTR<=5Mbs",   " DTR>5Mbs",   " DTR>10Mbs", " RotSpdTol>.5%",
+	" dStbOff",     " TrkOff",     " FmtGapReq", " nonMagnetic"
 };
 
-char *SlowMedFast[]	= {"slow", "medium", "fast", "eide"};
+char *ata_major_str[] = {
+	"reserved ",	"ATA-1 ",	"ATA-2 ",	"ATA-3 ",
+	"ATA-4 ",	"ATA-5 ",	"ATA-6 ",	"ATA-7 ",
+	"ATA-8 ",	"ATA-9 ",	"ATA-10 ",	"ATA-11 ",
+	"ATA-12 ",	"ATA-13 ",	"ATA-14 ",	"reserved "
+};
+
+char *ata_minor_str[] = {
+	"Reserved : ",						/* 0x0000 */
+	"ATA (ATA-1) X3T9.2 781D prior to revision 4 : ",	/* 0x0001 */
+	"ATA-1 published, ANSI X3.221-1994 : ",			/* 0x0002 */
+	"ATA (ATA-1) X3T10 781D revision 4 : ",			/* 0x0003 */
+	"ATA-2 published, ANSI X3.279-1996 : ",			/* 0x0004 */
+	"ATA-2 X3T10 948D prior to revision 2k : ",		/* 0x0005 */
+	"ATA-3 X3T10 2008D revision 1 : ",			/* 0x0006 */
+	"ATA-2 X3T10 948D revision 2k : ",			/* 0x0007 */
+	"ATA-3 X3T10 2008D revision 0 : ",			/* 0x0008 */
+	"ATA-2 X3T10 948D revision 3 : ",			/* 0x0009 */
+	"ATA-3 published, ANSI X3.298-199x : ",			/* 0x000A */
+	"ATA-3 X3T10 2008D revision 6 : ",			/* 0x000B */
+	"ATA-3 X3T13 2008D revision 7 and 7a : ",		/* 0x000C */
+	"ATA/ATAPI-4 X3T13 1153D revision 6 : ",		/* 0x000D */
+	"ATA/ATAPI-4 T13 1153D revision 13 : ",			/* 0x000E */
+	"ATA/ATAPI-4 X3T13 1153D revision 7 : ",		/* 0x000F */
+	"ATA/ATAPI-4 T13 1153D revision 18 : ",			/* 0x0010 */
+	"ATA/ATAPI-4 T13 1153D revision 15 : ",			/* 0x0011 */
+	"ATA/ATAPI-4 published, ANSI NCITS 317-1998 : ",	/* 0x0012 */
+	"ATA/ATAPI-5 T13 1321D revision 3 : ",			/* 0x0013 */
+	"ATA/ATAPI-4 T13 1153D revision 14 : ",			/* 0x0014 */
+	"ATA/ATAPI-5 T13 1321D revision 1 : ",			/* 0x0015 */
+	"Reserved : ",						/* 0x0016 */
+	"ATA/ATAPI-4 T13 1153D revision 17 : ",			/* 0x0017 */
+	"ATA/ATAPI-6 T13 1410D revision 0 : ",			/* 0x0018 */
+	"Reserved : "						/* 0x0019 */
+	"Reserved : "						/* 0x001A */
+	"Reserved : "						/* 0x001B */
+	"ATA/ATAPI-6 T13 1410D revision 1 : "			/* 0x001C */
+	"Reserved : "						/* 0x001D */
+	"Reserved : "						/* 0x001E */
+	"Reserved : "						/* 0x001F */
+};
+
+char *SlowMedFast[]	= {"slow", "medium", "fast", "eide", "ata"};
 char *BuffType[]	= {"unknown", "1Sect", "DualPort", "DualPortCache"};
 
 #define YN(b)	(((b)==0)?"no":"yes")
@@ -121,6 +172,7 @@ static void dump_identity (struct hd_driveid *id)
 {
 	int i;
 	char pmodes[64] = {0,}, dmodes[128]={0,};
+	unsigned short int *id_regs=(void*)id;
 
 	printf("\n Model=%.40s, FwRev=%.8s, SerialNo=%.20s", 
 		id->model, id->fw_rev, id->serial_no);
@@ -153,7 +205,13 @@ static void dump_identity (struct hd_driveid *id)
 	if (!(id->field_valid&1))
 		printf(" (maybe):");
 	printf(" CurCHS=%d/%d/%d, CurSects=%d",
-		id->cur_cyls, id->cur_heads, id->cur_sectors, *(int *)&id->cur_capacity0);
+		id->cur_cyls, id->cur_heads, id->cur_sectors,
+#ifdef BIG_ENDIAN
+		(id->cur_capacity0 << 16) | id->cur_capacity1
+#else
+		(id->cur_capacity1 << 16) | id->cur_capacity0
+#endif
+		);
 	printf(", LBA=%s", YN(id->capability&2));
 	if (id->capability&2)
  		printf(", LBAsects=%d", id->lba_capacity);
@@ -201,16 +259,22 @@ static void dump_identity (struct hd_driveid *id)
 				if (id->dma_ultra & 0x002)	strcat(dmodes,"udma1 ");
 				if (id->dma_ultra & 0x400)	strcat(dmodes,"*");
 				if (id->dma_ultra & 0x004)	strcat(dmodes,"udma2 ");
-				if (id->dma_ultra & 0x800)	strcat(dmodes,"*");
-				if (id->dma_ultra & 0x008)	strcat(dmodes,"udma3 ");
-				if (id->dma_ultra & 0x1000)	strcat(dmodes,"*");
-				if (id->dma_ultra & 0x0010)	strcat(dmodes,"udma4 ");
-				if (id->dma_ultra & 0x2000)	strcat(dmodes,"*");
-				if (id->dma_ultra & 0x0020)	strcat(dmodes,"udma5 ");
-				if (id->dma_ultra & 0x4000)	strcat(dmodes,"*");
-				if (id->dma_ultra & 0x0040)	strcat(dmodes,"udma6 ");
-				if (id->dma_ultra & 0x8000)	strcat(dmodes,"*");
-				if (id->dma_ultra & 0x0080)	strcat(dmodes,"udma7 ");
+#ifdef __NEW_HD_DRIVE_ID
+				if (id->hw_config & 0x2000) {
+#else /* !__NEW_HD_DRIVE_ID */
+				if (id->word93 & 0x2000) {
+#endif /* __NEW_HD_DRIVE_ID */
+					if (id->dma_ultra & 0x800)	strcat(dmodes,"*");
+					if (id->dma_ultra & 0x008)	strcat(dmodes,"udma3 ");
+					if (id->dma_ultra & 0x1000)	strcat(dmodes,"*");
+					if (id->dma_ultra & 0x0010)	strcat(dmodes,"udma4 ");
+					if (id->dma_ultra & 0x2000)	strcat(dmodes,"*");
+					if (id->dma_ultra & 0x0020)	strcat(dmodes,"udma5 ");
+					if (id->dma_ultra & 0x4000)	strcat(dmodes,"*");
+					if (id->dma_ultra & 0x0040)	strcat(dmodes,"udma6 ");
+					if (id->dma_ultra & 0x8000)	strcat(dmodes,"*");
+					if (id->dma_ultra & 0x0080)	strcat(dmodes,"udma7 ");
+				}
 			}
 #endif
 		}
@@ -218,7 +282,29 @@ static void dump_identity (struct hd_driveid *id)
 	if ((id->capability&1) && (id->field_valid&2))
 		printf(", tDMA={min:%d,rec:%d}", id->eide_dma_min, id->eide_dma_time);
 	printf("\n PIO modes: %s", pmodes);
-	printf("\n DMA modes: %s\n", dmodes);
+	printf("\n DMA modes: %s", dmodes);
+
+	printf("\n AdvancedPM=%s",YN(id_regs[83]&8));
+	if (id_regs[83] & 8) { 
+		if (!(id_regs[86]&8)) 
+			printf(": disabled (255)"); 
+		else if ((id_regs[91]&0xFF00)!=0x4000) 
+			printf(": unknown setting"); 
+		else 
+			printf(": mode=0x%02X (%d)",id_regs[91]&0xFF,id_regs[91]&0xFF); 
+	} 
+#ifdef __NEW_HD_DRIVE_ID
+	if ((id->minor_rev_num) || (id->major_rev_num)) {
+		printf("\n Drive Supports : ");
+		printf("%s", ata_minor_str[id->minor_rev_num]);
+		for (i=0; i<=15; i++) {
+			if (id->major_rev_num & (1<<i))
+				printf("%s", ata_major_str[i]);
+		}
+	}
+#endif /* __NEW_HD_DRIVE_ID */
+	printf("\n");	
+	printf("\n");
 }
 
 void flush_buffer_cache (int fd)
@@ -529,6 +615,10 @@ static void interpret_xfermode (unsigned int xfermode)
 }
 #endif /* HDIO_DRIVE_CMD */
 
+#ifndef VXVM_MAJOR
+#define VXVM_MAJOR 199
+#endif
+
 void process_dev (char *devname)
 {
 	int fd;
@@ -542,59 +632,63 @@ void process_dev (char *devname)
 		exit(errno);
 	}
 
+	switch(major(stat_buf.st_rdev)) {
 #ifdef SCSI_DISK0_MAJOR
-	if ((major(stat_buf.st_rdev) == SCSI_DISK0_MAJOR) 
-	 || (major(stat_buf.st_rdev) == SCSI_DISK1_MAJOR) 
-	 || (major(stat_buf.st_rdev) == SCSI_DISK2_MAJOR) 
-	 || (major(stat_buf.st_rdev) == SCSI_DISK3_MAJOR) 
-	 || (major(stat_buf.st_rdev) == SCSI_DISK4_MAJOR) 
-	 || (major(stat_buf.st_rdev) == SCSI_DISK5_MAJOR) 
-	 || (major(stat_buf.st_rdev) == SCSI_DISK6_MAJOR) 
-	 || (major(stat_buf.st_rdev) == SCSI_DISK7_MAJOR) 
+	case (SCSI_DISK0_MAJOR):
+	case (SCSI_DISK1_MAJOR):
+	case (SCSI_DISK2_MAJOR):
+	case (SCSI_DISK3_MAJOR):
+	case (SCSI_DISK4_MAJOR):
+	case (SCSI_DISK5_MAJOR):
+	case (SCSI_DISK6_MAJOR):
+	case (SCSI_DISK7_MAJOR):
 #else
-	if ((major(stat_buf.st_rdev) == SCSI_DISK_MAJOR) 
+	case (SCSI_DISK_MAJOR):
 #endif
 #ifdef MD_MAJOR
-	 || (major(stat_buf.st_rdev) == MD_MAJOR)
+	case (MD_MAJOR):
 #endif
-	) 
+	case (VXVM_MAJOR):
+	case (LVM_BLK_MAJOR):
 		is_scsi_hd = 1;
-	else
+		break;
 #ifdef XT_DISK_MAJOR
-	if (major(stat_buf.st_rdev) == XT_DISK_MAJOR) {
+	case (XT_DISK_MAJOR):
 		is_xt_hd = 1;
 		timing_MB = 2;
-	} else
+		break;
 #endif
-		if (major(stat_buf.st_rdev) != IDE0_MAJOR
-		 && major(stat_buf.st_rdev) != IDE1_MAJOR
+	case IDE0_MAJOR:
+	case IDE1_MAJOR:
 #ifdef IDE2_MAJOR
-		 && major(stat_buf.st_rdev) != IDE2_MAJOR
+	case IDE2_MAJOR:
 #endif
 #ifdef IDE3_MAJOR
-		 && major(stat_buf.st_rdev) != IDE3_MAJOR
+	case IDE3_MAJOR:
 #endif
 #ifdef IDE4_MAJOR
-		 && major(stat_buf.st_rdev) != IDE4_MAJOR
+	case IDE4_MAJOR:
 #endif
 #ifdef IDE5_MAJOR
-		 && major(stat_buf.st_rdev) != IDE5_MAJOR
+	case IDE5_MAJOR:
 #endif
 #ifdef IDE6_MAJOR
-		 && major(stat_buf.st_rdev) != IDE6_MAJOR
+	case IDE6_MAJOR:
 #endif
 #ifdef IDE7_MAJOR
-		 && major(stat_buf.st_rdev) != IDE7_MAJOR
+	case IDE7_MAJOR:
 #endif
 #ifdef IDE8_MAJOR
-		 && major(stat_buf.st_rdev) != IDE8_MAJOR
+	case IDE8_MAJOR:
 #endif
 #ifdef IDE9_MAJOR
-		 && major(stat_buf.st_rdev) != IDE9_MAJOR
+	case IDE9_MAJOR:
 #endif
-	) {
+		break;	/* do nothing */
+	default:
 		fprintf(stderr,"%s not supported by hdparm\n",devname);
 		exit(EINVAL);
+		break;
 	}
 
 	fd = open (devname, O_RDONLY|O_NONBLOCK);
@@ -748,7 +842,7 @@ void process_dev (char *devname)
 		unsigned char args[4] = {WIN_SETFEATURES,0,0x04,0};
 		no_scsi();
 		args[2] = defects ? 0x04 : 0x84;
-		if (get_prefetch)
+		if (get_defects)
 			printf(" setting drive defect-mgmt to %ld\n", defects);
 		if (ioctl(fd, HDIO_DRIVE_CMD, &args))
 			perror(" HDIO_DRIVE_CMD(defectmgmt) failed");
@@ -787,6 +881,26 @@ void process_dev (char *devname)
 		if (ioctl(fd, HDIO_DRIVE_CMD, &args))
 			perror(" HDIO_DRIVE_CMD(setreadahead) failed");
 	}
+	if (set_apmmode) {
+		unsigned char args[4] = {WIN_SETFEATURES,0,0,0};
+		no_scsi();
+		if (apmmode<1) apmmode=1;
+		if (apmmode>255) apmmode=255;
+		if (get_apmmode) 
+			printf(" setting Advanced Power Management level to");
+			if (apmmode==255) { 
+			/* disable Advanced Power Management */
+				args[2] = 0x85; /* feature register */ 
+				if (get_apmmode) printf(" disabled\n");
+			} else {
+				/* set Advanced Power Management mode */
+				args[2] = 0x05; /* feature register */  
+				args[1] = apmmode; /* sector count register */
+				if (get_apmmode) printf(" 0x%02lX (%ld)\n",apmmode,apmmode);
+			} 
+			if (ioctl(fd, HDIO_DRIVE_CMD, &args))
+				perror(" HDIO_DRIVE_CMD failed");
+	}
 #ifdef CDROM_SELECT_SPEED
 	if (set_cdromspeed) {
 		printf ("setting cdrom speed to %ld\n", cdromspeed);
@@ -979,11 +1093,20 @@ void process_dev (char *devname)
 		}
 	}
 	if (verbose || get_geom) {
+#if HDIO_GETGEO_BIG
+		static struct hd_big_geometry g;
+#else
 		static struct hd_geometry g;
+#endif
 		if (ioctl(fd, BLKGETSIZE, &parm))
 			perror(" BLKGETSIZE failed");
+#if HDIO_GETGEO_BIG
+		else if (ioctl(fd, HDIO_GETGEO_BIG, &g))
+			perror(" HDIO_GETGEO_BIG failed");
+#else
 		else if (ioctl(fd, HDIO_GETGEO, &g))
 			perror(" HDIO_GETGEO failed");
+#endif
 		else	printf(" geometry     = %d/%d/%d, sectors = %ld, start = %ld\n",
 				g.cylinders, g.heads, g.sectors, parm, g.start);
 	}
@@ -1011,6 +1134,23 @@ void process_dev (char *devname)
 		printf(" drive state is:  %s\n", state);
 	}
 #endif
+#ifdef HDIO_DRIVE_RESET
+	if (perform_reset) {
+		no_scsi();
+		no_xt();
+		if (ioctl(fd, HDIO_DRIVE_RESET, NULL))
+			 perror(" HDIO_DRIVE_RESET failed");
+	}
+#endif /* HDIO_DRIVE_RESET */
+#ifdef HDIO_TRISTATE_HWIF
+	if (perform_tristate) {
+		unsigned char args[4] = {0,tristate,0,0};
+		no_scsi();
+		no_xt();
+		if (ioctl(fd, HDIO_TRISTATE_HWIF, &args))
+			perror(" HDIO_TRISTATE_HWIF failed");
+	}
+#endif /* HDIO_TRISTATE_HWIF */
 	if (get_identity) {
 		no_scsi();
 		no_xt();
@@ -1067,6 +1207,7 @@ void usage_error (void)
 #ifdef HDIO_DRIVE_CMD
 	" -A   set drive read-lookahead flag (0/1)\n"
 #endif
+	" -B   get Advanced Power Management setting (1-255)\n"
 #ifdef HDIO_SET_32BIT
 	" -c   get/set IDE 32-bit IO setting\n"
 #endif
@@ -1121,7 +1262,13 @@ void usage_error (void)
 	" -v   default; same as -acdgkmnru (-gr for SCSI, -adgr for XT)\n"
 	" -V   display program version and exit immediately\n"
 #ifdef HDIO_DRIVE_CMD
+#ifdef HDIO_DRIVE_RESET
+	" -w   perform device reset (DANGEROUS)\n"
+#endif
 	" -W   set drive write-caching flag (0/1) (DANGEROUS)\n"
+#ifdef HDIO_TRISTATE_HWIF
+	" -x   perform device for hotswap flag (0/1) (DANGEROUS)\n"
+#endif
 	" -X   set IDE xfer mode (DANGEROUS)\n"
 	" -y   put IDE drive in standby mode\n"
 	" -Y   put IDE drive to sleep\n"
@@ -1263,7 +1410,7 @@ int main(int argc, char **argv)
 						break;
 
 					case 'D':
-						set_defects = noisy;
+						get_defects = noisy;
 						noisy = 1;
 						GET_NUMBER(set_defects,defects);
 						if (!set_defects)
@@ -1426,11 +1573,34 @@ int main(int argc, char **argv)
 						scan_hwif = 1;
 						break;
 #endif /* HDIO_SCAN_HWIF */
+#ifdef HDIO_DRIVE_RESET
+					case 'w':
+						perform_reset = 1;
+						break;
+#endif /* HDIO_DRIVE_RESET */
+#ifdef HDIO_TRISTATE_HWIF
+					case 'x':
+						if (!*p && argc && isdigit(**argv))
+							p = *argv++, --argc;
+						if (*p == '0' || *p == '1') {
+							perform_tristate = 1;
+							tristate = *p++ - '0';
+						} else
+							fprintf(stderr, "-x: missing value (0/1)\n");
+						break;
+
+#endif /* HDIO_TRISTATE_HWIF */
 					case 'a':
 						get_readahead = noisy;
 						noisy = 1;
 						GET_NUMBER(set_readahead,readahead);
 						break;
+					case 'B':
+						noisy = 1;
+						GET_NUMBER(set_apmmode,apmmode);
+						if (!set_apmmode)
+						printf("-B: missing value (1-255)\n");
+						break;
 					case 't':
 						do_timings = 1;
 						do_flush = 1;
